#pragma once

#include <map>
#include <memory>

#include <dune/grid/common/mcmgmapper.hh>

#include <amdis/Output.hpp>
#include <amdis/typetree/TreeContainer.hpp>

namespace AMDiS
{
  /**
   * \addtogroup Adaption
   * @{
   **/

  enum class DataTransferOperation {
    NO_OPERATION = 0,
    INTERPOLATE  = 1
  };


  /// \brief Interface for Containers allowing data transfer between grid changes.
  template <class Container>
  class DataTransferInterface
  {
  public:
    /// Virtual destructor
    virtual ~DataTransferInterface() = default;

    /// Collect data that is needed before grid adaption
    virtual void preAdapt(Container const& container, bool mightCoarsen) = 0;

    /// Interpolate data to new grid after grid adaption
    virtual void postAdapt(Container& container, bool refined) = 0;
  };


  /**
   * \brief Implementation of \ref DataTransferInterface that does not interpolate, but
   * just resizes the containers to the dimension of the basis
   **/
  template <class Container>
  class NoDataTransfer
      : public DataTransferInterface<Container>
  {
  public:
    void preAdapt(Container const& container, bool) override {}

    void postAdapt(Container& container, bool) override
    {
      container.resize();
    }
  };


  template <class Node, class Container, class Basis>
  class NodeDataTransfer;


  /** Data Transfer implementation for a single grid using interpolation
   *  Handles computations related to the geometric information of the grid and passes that to the
   *    underlying NodeDataTransfer classes
   */
  template <class Container, class Basis>
  class DataTransfer
      : public DataTransferInterface<Container>
  {
    using LocalView = typename Basis::LocalView;
    using Tree = typename LocalView::Tree;
    using GridView = typename Basis::GridView;
    using Grid = typename GridView::Grid;

    using Element = typename GridView::template Codim<0>::Entity;
    using Geometry = typename Element::Geometry;
    using LocalCoordinate = typename Geometry::LocalCoordinate;
    using IdType = typename Grid::LocalIdSet::IdType;

    template <class Node>
    using NodeElementData = typename NodeDataTransfer<Node, Container, Basis>::NodeElementData;
    using ElementData = TYPEOF(makeTreeContainer<NodeElementData>(std::declval<const Tree&>()));

  public:
    DataTransfer(std::shared_ptr<Basis> basis);

    /** Saves data contained in coeff in the PersistentContainer
     *  To be called after grid.preAdapt() and before grid.adapt()
     */
    void preAdapt(Container const& coeff, bool mightCoarsen) override;

    /** Unpacks data from the PersistentContainer
     *  To be called after grid.adapt() and before grid.postAdapt()
     */
    // [[expects: basis is update in gridView]]
    // [[expects: comm is updated in basis]]
    void postAdapt(Container& coeff, bool refined) override;

  private:
    /// The global basis
    std::shared_ptr<Basis const> basis_;

    /// Container with data that persists during grid adaptation
    using PersistentContainer = std::map<IdType, ElementData>;
    PersistentContainer persistentContainer_;

    /// Map leaf entities to unique index
    using Mapper = Dune::LeafMultipleCodimMultipleGeomTypeMapper<Grid>;
    Mapper mapper_;

    /// Data transfer on a single basis node
    template <class Node>
    using NDT = NodeDataTransfer<Node, Container, Basis>;
    using NodeDataTransferContainer = TYPEOF(makeTreeContainer<NDT>(std::declval<const Tree&>()));
    NodeDataTransferContainer nodeDataTransfer_;
  };


  /// Factory to create DataTransfer objects based on the \ref DataTransferOperation
  template <class Container>
  class DataTransferFactory
  {
    using Interface = DataTransferInterface<Container>;

  public:
    template <class Basis>
    static std::unique_ptr<Interface> create(DataTransferOperation op, std::shared_ptr<Basis> basis)
    {
      switch (op)
      {
        case DataTransferOperation::NO_OPERATION:
          return std::make_unique<NoDataTransfer<Container>>();
        case DataTransferOperation::INTERPOLATE:
          return std::make_unique<DataTransfer<Container, Basis>>(std::move(basis));
        default:
          error_exit("Invalid data transfer\n");
          return nullptr; // avoid warnings
      }
    }
  };

  /// @}

} // end namespace AMDiS

#include "DataTransfer.inc.hpp"
