#pragma once

#include <functional>
#include <list>

#include <amdis/AdaptionInterface.hpp>
#include <amdis/Output.hpp>

namespace AMDiS
{
  /**
   * \addtogroup Adaption
   * @{
   **/

  /// \brief Implementation of \ref AdaptionInterface for concrete Grid type.
  template <class Grid>
  class GridTransfer
      : public AdaptionInterface
  {
    using Self = GridTransfer;
    using Key = std::uintptr_t;

    struct InterpolateCallback
    {
      std::function<void(bool)> preAdapt;
      std::function<void(bool)> postAdapt;
      int count = 0;
    };

    struct UpdateCallback
    {
      std::function<void(void)> update;
      int count = 0;
    };

  public:
    /// Bind a (mutable) grid to this GridTransfer. Must be called before any xxxAdapt() method.
    // [[ensures: bound()]]
    void bind(Grid& grid)
    {
      grid_ = &grid;
    }

    /// Release the grid from this GridTransfer
    // [[ensures: not bound()]]
    void unbind()
    {
      grid_ = nullptr;
    }

    /// Returns whether this GridTransfer is bound to a grid.
    bool bound() const
    {
      return grid_ != nullptr;
    }

    /// Register data to GridTransfer.
    /**
     * This inserts a callback that is called in \ref preAdapt() and another that is
     * called in \ref postAdapt() step.
     *
     * *Requirement:*
     * - Data must be a model of \ref Concepts::InterpolateData
     **/
    template <class Data,
      REQUIRES(Concepts::InterpolateData<Data>)>
    void attach(Data& data)
    {
      auto it = interpolateCallbacks_.emplace(Key(&data), InterpolateCallback{
        [&data](bool mightCoarsen) -> void { data.preAdapt(mightCoarsen); },
        [&data](bool refined) -> void { data.postAdapt(refined); }});

      it.first->second.count++;
    }

    /// Register data to GridTransfer.
    /**
     * This inserts a callback that is called after the \ref adapt() step.
     * The callback calls `data.update(data.gridView())`.
     *
     * *Requirement:*
     * - Data must be a model of \ref Concepts::UpdateData
     **/
    template <class Data,
      REQUIRES(Concepts::UpdateData<Data>)>
    void attach(Data& data)
    {
      attach(data, [&data]() -> void { data.update(data.gridView()); });
    }

    /// Register callback to GridTransfer
    template <class Data, class Callback>
    void attach(Data const& data, Callback&& callback)
    {
      auto it = updateCallbacks_.emplace(Key(&data), UpdateCallback{FWD(callback)});
      it.first->second.count++;
    }

    /// Unregister data from GridTransfer. Data is identified by its address.
    template <class Data,
      REQUIRES(Concepts::InterpolateData<Data>)>
    void detach(Data const& data)
    {
      eraseCallback(interpolateCallbacks_, Key(&data));
    }

    /// Unregister data from GridTransfer. Data is identified by its address.
    template <class Data,
      REQUIRES(Concepts::UpdateData<Data>)>
    void detach(Data const& data)
    {
      eraseCallback(updateCallbacks_, Key(&data));
    }

    /// Implements \ref GridTransferInterface::preAdapt
    // [[expects: bound()]]
    bool preAdapt() override
    {
      assert(bound());
      Dune::Timer t;
      mightCoarsen_ = grid_->preAdapt();
      Dune::MPIHelper::getCollectiveCommunication().max(&mightCoarsen_, 1);

      for (auto&& data : interpolateCallbacks_)
        data.second.preAdapt(mightCoarsen_);

      info(2,"preAdapt needed {} seconds", t.elapsed());
      return mightCoarsen_;
    }

    /// Implements \ref GridTransferInterface::adapt
    // [[expects: bound()]]
    bool adapt() override
    {
      assert(bound());
      Dune::Timer t;
      refined_ = grid_->adapt();
      Dune::MPIHelper::getCollectiveCommunication().max(&refined_, 1);

      info(2,"adapt needed {} seconds", t.elapsed());
      return refined_;
    }

    /// Implements \ref GridTransferInterface::postAdapt
    // [[expects: bound()]]
    void postAdapt() override
    {
      assert(bound());
      Dune::Timer t;
      if (mightCoarsen_ || refined_) {
        for (auto&& data : updateCallbacks_)
          data.second.update();
        for (auto&& data : interpolateCallbacks_)
          data.second.postAdapt(refined_);
      }

      grid_->postAdapt();
      changeIndex_++;
      info(2,"postAdapt needed {} seconds", t.elapsed());
    }

    /// Returns the grid change index, see \ref changeIndex.
    unsigned long changeIndex() const
    {
      return changeIndex_;
    }

  private:
    // remove entry from map if count <= 1, otherwise reduce count by 1
    template <class Map>
    void eraseCallback(Map& map, Key key)
    {
      auto it = map.find(key);
      if (it == map.end())
        warning("Data to detach not found");
      else if (it->second.count > 1)
        it->second.count--;
      else
        map.erase(it);
    }

  private:
    /// A grid this GridTransfer is bound to in \ref bind()
    Grid* grid_ = nullptr;

    /// A list of data containers handled during grid adaption
    std::map<Key,InterpolateCallback> interpolateCallbacks_;
    std::map<Key,UpdateCallback> updateCallbacks_;

    /// Flag set during \ref preAdapt(), indicating whether any element might be coarsened in \ref adapt()
    bool mightCoarsen_ = false;

    /// Flag set during \ref adapt() indicating that at least one entity was refined
    bool refined_ = false;

    /// This index is incremented every time the grid is changed, e.g. by refinement or coarsening.
    unsigned long changeIndex_ = 0;
  };

  /// @}

} // end namespace AMDiS
